OBJ
  'eeprom : "Basic_I2C_Driver"
  eeprom : "pasm_i2c_driver"
  'eeprom_p : "pasm_i2c_driver_p"
  ser : "cserial"
{{
  This object implements an EEPROM file system.  A sector size of 64
  is used, and the EEPROM must have a page size of 64 bytes or greater.
  Files consist of multiple sectors that a daisy-chained together.
  The first long of each sector points to the next sector in the
  chain.

  When the file system is first created the least significant
  bit of first long of each sector is set to one to indicate that
  the sector is unused.  When a sector is allocated for a file the
  last six bits in the first long are cleared to indicate that it
  is in use.

  The first long in the first sector contains the ASCII characters
  "EEFS" to indicate that this is a valid EEPROM file system.  The
  next long contains the address of the root directory, which is
  normally the second sector in the file system.  The third long
  contains the size of the root directory in bytes.

  In general, directory entries are 30 bytes in size.  They consist
  of 4 bytes for the starting address and attribute bits, 4 bytes
  for the file size and 22 bytes for the file name.  There are two
  30-byte directory entries per sector.  Since the file starting
  address is a multiple of 64, the 6 least significant bit are
  zero.  These bits are used to store the attibutes for the file.
  The attributes bits are directory, read, write and execute, or
  "drwx".

  A fifth bit is set to indicate that the directory entry is active
  and does not represent a deleted file.  If all 6 of the least
  significant bits are zero, the directory entry is inactive, and
  it is available to be used for a new file.

  A 64-byte cache is used to hold the contents of the current sector
  that is being read or written.  Two flags are used to indicate
  the range of bytes that have changed in the sector.  These are
  called dirtylow and dirtyhigh.
  
}}
  
CON
' file system constants
  magic_word      = $45454653 ' First word of a valid file system
  'disk_start      = $08000    ' Starting address in the EEPROM

  ' 32Kx8 EEPROM
  'disk_start      = $0040     ' Starting address in the EEPROM
'  disk_start      = $2a00     ' Starting address in the EEPROM
'  disk_end        = $8000     ' Address after the end of the file system

  {
  ' 64Kx8 EEPROM
  disk_start      = $08000    ' Starting address in the EEPROM
  disk_end        = $10000    ' Address after the end of the file system
  }
{
  ' 128Kx8 EEPROM
  disk_start      = $08000    ' Starting address in the EEPROM
  disk_end        = $20000    ' Address after the end of the file system
}
  sector_size     = 64        ' Number of bytes in a sector
  device_sel      = $a0       ' EEPROM I2C Device Select
  attrib_mask     = $3f       ' Mask for the file attribute bits
  sector_mask     = $ffffffc0 ' Mask for the starting address of a sector
  dirent_size     = 30        ' Number of bytes in a directory entry
  fname_size      = 22        ' Maximum number of characters in a file name

' byte offsets for handle variables
  s_sectoraddr    =  0        ' Current sector address
  s_offset        =  4        ' Current offset within a sector 
  s_position      =  8        ' Current byte position within a file
  s_filesize      = 12        ' Number of bytes in the file
  s_cachedaddr    = 16        ' Cached sector address
  s_cacheptr      = 20        ' Pointer to the memory cache
  s_dirtylow      = 24        ' Low offset for a changed value
  s_dirtyhigh     = 28        ' High offset for a changed value
  s_firstaddr     = 32        ' First sector address including attribute bits
  s_directaddr    = 36        ' Address of the file's directory entry

  handle_size     = 40        ' Size of the handle struct in bytes
  handle_longs    = 10        ' Size of the handle struct in longs

DAT
  disk_start long $8000
  disk_end long 0
  eeprom_pin long 0 ' First pin of the EEPROM
  first_free long disk_start + sector_size
  local_handle long 0[handle_longs]
  local_buffer long 0[sector_size/4]
  readdir_block long 0[8] 'attrib, size, fname

'****************************************************************
' Standard FSRW Methods
'****************************************************************
PUB mount_explicit(clkpin, datpin, startaddr, size)
{{
  This routine initializes the disk for file access.  It returns
  a zero if the disk is ready, or a negative error code if it is not.
}}
  
  eeprom_pin := clkpin
  disk_start := startaddr
  eeprom.initialize(eeprom_pin)
  if (eeprom.ReadLong(eeprom_pin, device_sel, disk_start) <> magic_word)
    ser.dbprintf0(string("EEPROM is not formatted.  Formating...\n"))
    format(size)
    'return -20 ' Signal mount failure
  disk_end := eeprom.ReadLong(eeprom_pin, device_sel, disk_start + 12)
  InitRootHandle(@local_handle)

PUB pflush
  result := hflush(@local_handle)

PUB pclose
  result := hclose(@local_handle)

PUB popen(fname, mode)
  hclose(@local_handle)
  InitRootHandle(@local_handle)
  result := hopen(@local_handle, fname, mode)

PUB get_filesize
  result := hget_filesize(@local_handle)

PUB pread(buffer, num)
  result := hread(@local_handle, buffer, num)

PUB pgetc | val
  if (result := hread(@local_handle, @val, 1)) == 1
    result := val & $ff

PUB pwrite(buffer, num)
  result := hwrite(@local_handle, buffer, num)

PUB pputs(buffer)
  result := hwrite(@local_handle, buffer, strsize(buffer))

PUB pputc(val)
  if (result := hwrite(@local_handle, @val, 1)) == 1
    result := 0

PUB seek(position)
  result := hseek(@local_handle, position)

PUB tell
  result := htell(@local_handle)

PUB opendir
  hclose(@local_handle)
  InitRootHandle(@local_handle)

PUB nextfile(fbuf)
  result := hnextfile(@local_handle, fbuf)

'===================================================================
' Extended FSRW methods that use a file handle
'===================================================================
PUB hopen(handle, fname, mode)
{{
  This routine opens a file for access.  The access mode
  is determined by the first character of the string "mode".
  The allowable values for the first character are "r", "w"
  and "a" for read, write and append respectively.  The
  pass value of the handle pointer is return if successful.
}}
  if (mode == "d")
    result := Remove(fname)
  elseif (handle == 0)
    return -2
  elseif (mode == "r")
    result := OpenRead(fname, handle)
  elseif (mode == "w")
    result := OpenWrite(fname, handle)
  elseif (mode == "a")
    result := OpenAppend(fname, handle)
  ifnot result
    result := -1
  else
    result :=  0

PUB hclose(handle) | position, filesize
  position := long[handle + s_position]
  filesize := long[handle + s_filesize]
  hflush(handle)
  long[handle + s_cachedaddr] := -1
  UpdateFilesize(handle, position, filesize)

PUB hread(handle, buffer, num) | sectoraddr, offset, position, filesize, cachedaddr, cacheptr, num1, i
{{
  This routine reads "num" bytes from the file defined by "handle".
  The bytes are copies into the memory pointed to by "buffer".
  The return value consist of the number of bytes that were actually
  read.  A zero return value indicates an end of file or an error
  condition.
}}
  longmove(@sectoraddr, handle, 6) ' Copy first 6 variables from the handle
  i := 0
  repeat while (i < num AND position < filesize)
    if (offset => sector_size)
      if cachedaddr <> sectoraddr
        sectoraddr := eeprom.ReadLong(eeprom_pin, device_sel, sectoraddr)
      else
        sectoraddr := long[cacheptr]
      offset := 4
    if cachedaddr <> sectoraddr
      eeprom.ReadPage(eeprom_pin, device_sel, sectoraddr, cacheptr, sector_size)
      cachedaddr := sectoraddr
    num1 := (num - i) <# (filesize - position) ' Get smallest value of num - i,
    num1 := num1 <# (sector_size - offset)     ' filesize - position, and sector_size - offset
    bytemove(buffer + i, cacheptr + offset, num1)
    position += num1
    offset += num1
    i += num1
  longmove(handle, @sectoraddr, 5) ' Update the first 5 variables in the handle
  return i

PUB hwrite(handle, buffer, num) | sectoraddr, offset, position, filesize, cachedaddr, cacheptr, dirtylow, dirtyhigh, firstaddr, directaddr, num1, nextaddr
{{
  This routine writes "num" bytes to the file defined by "handle".
  The bytes are copied from the memory pointed to by "buffer".
  The return value consist of the number of bytes that were actually
  written.  A zero return value indicates an error occurred.
}}
  'ser.dbprintf1(string("hwrite: %d\n"), num)
  if (handle == 0)
    return -1
  if (num < 1)
    return 0
  result := num
  longmove(@sectoraddr, handle, 10) ' Copy the 10 variables from the handle
  'ser.dbprintf4(string("%4x, %d, %d, %d\n"), sectoraddr, offset, position, filesize)
  if (sectoraddr == 0)
    sectoraddr := FindFreeSector
    firstaddr |= sectoraddr
    WriteEEPromLong(directaddr, firstaddr)  
    cachedaddr := sectoraddr
    long[cacheptr] := 0
    dirtylow := 0
    dirtyhigh := 4
  elseif cachedaddr <> sectoraddr
    WriteDirtyBlock(cachedaddr, cacheptr, @dirtylow, @dirtyhigh)
    eeprom.ReadPage(eeprom_pin, device_sel, sectoraddr, cacheptr, sector_size)
    cachedaddr := sectoraddr
  repeat
    num1 := num <# (sector_size - offset)
    bytemove(cacheptr + offset, buffer, num1)
    dirtylow <#= offset ' Update dirtylow if offset is smaller
    num -= num1
    position += num1
    offset += num1
    dirtyhigh #>= offset ' Update dirtyhigh if offset is larger
    if (num =< 0)
      quit
    buffer += num1
    nextaddr := long[cacheptr]
    if nextaddr
      WriteDirtyBlock(sectoraddr, cacheptr, @dirtylow, @dirtyhigh)
      eeprom.ReadPage(eeprom_pin, device_sel, nextaddr, cacheptr, sector_size)
    else
      nextaddr := FindFreeSector
      if (nextaddr == 0)
        return 0
      long[cacheptr] := nextaddr
      dirtylow := 0
      dirtyhigh #>= 4 ' Update dirtyhigh if less than 4
      WriteDirtyBlock(sectoraddr, cacheptr, @dirtylow, @dirtyhigh)
      long[cacheptr] := 0
      dirtylow := 0
      dirtyhigh := 4
    sectoraddr := nextaddr
    cachedaddr := sectoraddr
    offset := 4
  ifnot dirtylow ' Write if sectoraddr has changed
    WriteDirtyBlock(sectoraddr, cacheptr, @dirtylow, @dirtyhigh)
  longmove(handle, @sectoraddr, 9) ' Update the first 9 variables in the handle
  
PUB hgetc(handle) | val
  if (result := hread(handle, @val, 1)) == 1
    result := val & $ff

PUB hputs(handle, buffer)
  result := hwrite(handle, buffer, strsize(buffer))

PUB hputc(handle, val)
  if (result := hwrite(handle, @val, 1)) == 1
    result := 0

PUB hopendir(handle) | directaddr, firstaddr, filesize, cacheptr
{{
  This routine opens the current directory for reading by readdir.
  A handle is returned, which is used by subsequent calls by readdir.
}}
  directaddr := long[@local_handle + s_directaddr]
  firstaddr  := long[@local_handle + s_firstaddr]
  filesize   := long[@local_handle + s_filesize]
  cacheptr   := handle + handle_size
  InitFileHandle(handle, directaddr, firstaddr, filesize, cacheptr)

PUB hnextfile(handle, fbuf)
  result := hreaddir(handle)
  if result
    bytemove(fbuf, result + 8, strsize(result + 8) + 1)

PUB hget_filesize(handle)
{{
  This routine returns the size of the file in bytes.
}}
  ifnot handle
    return
  result := long[handle + s_filesize]

PUB hseek(handle, newpos) | sectoraddr, offset, position, filesize, num
{{
  This routine moves to the location specified by "newpos".
}}
  ifnot handle
    return

  longmove(@sectoraddr, handle, 4) ' Copy first 4 variables from the handle

  if (newpos < 0) or (newpos > filesize)
    return 0

  ' Check if the filesize needs to be updated before changing position
  UpdateFilesize(handle, position, filesize)

  ' Backup to the beginning of the current block
  position -= offset - 4
  
  ' Backup to the beginning of the file if newpos < position
  if newpos < position
    sectoraddr := long[handle + s_firstaddr] & sector_mask
    position := 0

  ' Advance to the new position
  offset := newpos - position + 4
  position := newpos
  repeat while offset > sector_size
    sectoraddr := eeprom.ReadLong(eeprom_pin, device_sel, sectoraddr)
    offset -= sector_size - 4

  'ser.dbprintf3(string("seek: %4x, %d, %d\n"), sectoraddr, offset, position)

  longmove(handle, @sectoraddr, 3) ' Copy the first 3 variables back to the handle

  return 1

PUB htell(handle)
{{
  This routine returns the current position from the beginning of the file.
}}
  ifnot handle
    return
  result := long[handle + s_position]

PUB hflush(handle) | sectoraddr, cachedaddr, cacheptr, dirtylow, dirtyhigh
{{
  This routine writes out the data that is sitting in the write buffer.
}}
  ifnot handle
    return
  sectoraddr := long[handle + s_sectoraddr]
  cachedaddr := long[handle + s_cachedaddr]
  cacheptr := long[handle + s_cacheptr]
  dirtylow := long[handle + s_dirtylow]
  dirtyhigh := long[handle + s_dirtyhigh]
  WriteDirtyBlock(sectoraddr, cacheptr, @dirtylow, @dirtyhigh)
  long[handle + s_dirtylow] := dirtylow
  long[handle + s_dirtyhigh] := dirtyhigh

'=============================================================================
' New methods
'=============================================================================
PUB handlesize
{{
  This routine returns the size of a file handle in bytes so that
  the calling routine can allocate a handle that is used by most
  of the file access routines.
}}
  return handle_size + sector_size

PUB format(size) | sectoraddr
{{
  This routine formats the disk with an empty file system.
}}
  disk_end := size
  WriteEEPromLong(disk_start, magic_word)
  sectoraddr := disk_start + sector_size
  WriteEEPromLong(disk_start + 4, sectoraddr | 1)
  WriteEEPromLong(disk_start + 8, 0)
  WriteEEPromLong(disk_start + 12, disk_end)
  WriteEEPromLong(sectoraddr, 0)
  sectoraddr += sector_size
  first_free := sectoraddr
  repeat while (sectoraddr < disk_end)
    WriteEEPromLong(sectoraddr, 1)
    sectoraddr += sector_size
  InitRootHandle(@local_handle)
  
PUB rename(fname1, fname2) | filesize
{{
  This routine renames a file in the files system.  It returns
  a value of zero if the file did not exist, or a one if it did.
}}
  if FindFile(fname2, @local_handle, @filesize)
    return -1
  ifnot FindFile(fname1, @local_handle, @filesize)
    return
  long[@local_handle + s_position] -= fname_size
  long[@local_handle + s_offset] -= fname_size
  hwrite(@local_handle, fname2, fname_size)
  hflush(@local_handle)
  return 1

PUB hreaddir(dirhand) | sectoraddr, filesize, filename
{{
  This routine reads the information for the next file in the
  directory indicated by the handle "dirhand".  It returns a
  pointer to the file directory information block if the end of
  the directory has not been reached.  Otherwise it will return
  a zero.  The file directory information block contains 3 longs
  with the filesize, attributes, a physical size followed by the
  file name.
}}
  if dirhand == 0
    return
  repeat
    if hread(dirhand, @sectoraddr, 4) <> 4
      return
    if hread(dirhand, @filesize, 4) <> 4
      return
    if hread(dirhand, @readdir_block[2], fname_size) <> fname_size
      return
    'quit 'Uncomment this line to display deleted files
    if sectoraddr & attrib_mask
      quit

  ifnot sectoraddr & attrib_mask
    readdir_block[0] := 0
    readdir_block[1] := 0
  else
    readdir_block[0] := sectoraddr
    readdir_block[1] := filesize
  return @readdir_block

PUB hclosedir(dirhand)
{{
  This routines closes the directory that had previously been
  opened by opendir.
}}

PUB chdir(path) | directaddr, sectoraddr, filesize, nextptr, currentwd_save[handle_longs]
{{
  This routine changes the current working directory to the path indicated by
  "path".  A value of zero is returned on success, and minus one is returned if the
  path does not exist.
}}
  'ser.dbprintf1(string("EEFS: chdir %s\n"), path)
  longmove(@currentwd_save, @local_handle, handle_longs) ' Save CWD

  ' Check if starting from root directory
  if byte[path] == "/"
    path++
    InitRootHandle(@local_handle)

  ' Loop over sub-directory names in path
  repeat while byte[path]
    nextptr := path
    repeat while byte[nextptr]
      if byte[nextptr++] == "/"
        byte[nextptr-1] := 0
        quit
    'ser.dbprintf1(string("FindFile %s\n"), path)
    sectoraddr := FindFile(path, @local_handle, @filesize)
    ifnot sectoraddr ' If not file restore cwd
      longmove(@local_handle, @currentwd_save, handle_longs)
      long[@local_handle + s_cachedaddr] := -1
      return -1
    directaddr := long[@local_handle + s_sectoraddr] + long[@local_handle + s_offset] - dirent_size
    InitFileHandle(@local_handle, directaddr, sectoraddr, filesize, @local_buffer)
    if byte[nextptr]
      byte[nextptr-1] := "/"
    path := nextptr
  return 0

PUB mkdir(fname, mode)
  return CreateFile(fname, %1111) - 1

PUB chmod(fname, modebits) | sectoraddr, filesize
  ifnot (sectoraddr := FindFile(fname, @local_handle, @filesize))
    return
  long[@local_handle + s_position] -= dirent_size
  long[@local_handle + s_offset] -= dirent_size
  sectoraddr := (sectoraddr & sector_mask) | %10000 | (modebits & %1111)
  hwrite(@local_handle, @sectoraddr, 4)
  hflush(@local_handle)

'****************************************************************
' Private Methods
'****************************************************************
PRI UpdateFilesize(handle, position, filesize) | directaddr
  if position > filesize
    filesize := position
    long[handle + s_filesize] := filesize
    directaddr := long[handle + s_directaddr]
    WriteEEPromLong(directaddr + 4, filesize)

PRI WriteDirtyBlock(sectoraddr, buffer, pdirtylow, pdirtyhigh) | dirtylow, dirtyhigh
  dirtylow := long[pdirtylow]
  dirtyhigh := long[pdirtyhigh]
  if dirtylow < 0 or dirtyhigh > sector_size
    'ser.dbprintf2(string("WriteDirtyBlock: ERROR - dirtylow = %d, dirtyhigh = %d"), dirtylow, dirtyhigh)
    return
  if dirtylow => dirtyhigh
    return
  eeprom.WritePage(eeprom_pin, device_sel, sectoraddr + dirtylow, buffer + dirtylow, dirtyhigh - dirtylow)
  repeat while(eeprom.WriteWait(eeprom_pin, device_sel, sectoraddr + dirtylow))
  long[pdirtylow] := sector_size
  long[pdirtyhigh] := 0

PRI ClearCached(handle)
  long[handle + s_cachedaddr] := -1
  long[handle + s_dirtylow]   := sector_size
  long[handle + s_dirtyhigh]  := 0
  
PRI InitFileHandle(handle, directaddr, firstaddr, filesize, cacheptr)
{{
  This routine intializes a file handle with the directory location
  and file information for a file.  The handle is initialized for
  the beginning of the file.
}}
  'ser.dbprintf2(string("InitFileHandle: %4x, %d\n"), firstaddr, filesize)
  long[handle + s_directaddr] := directaddr
  long[handle + s_firstaddr]  := firstaddr
  long[handle + s_sectoraddr] := firstaddr & sector_mask
  long[handle + s_offset]     := 4
  long[handle + s_filesize]   := filesize
  long[handle + s_position]   := 0
  long[handle + s_cachedaddr] := -1
  long[handle + s_cacheptr]   := cacheptr
  long[handle + s_dirtylow]   := sector_size
  long[handle + s_dirtyhigh]  := 0
  return handle

PRI InitRootHandle(handle) | firstaddr, filesize
{{
  This routine initializes a file handle for to point to the beginning
  of the root directory.
}}
  firstaddr := eeprom.ReadLong(eeprom_pin, device_sel, disk_start + 4) ' Get root directory address
  filesize  := eeprom.ReadLong(eeprom_pin, device_sel, disk_start + 8) ' Get root directory size
  InitFileHandle(handle, disk_start + 4, firstaddr, filesize, @local_buffer)
  return handle

' This routine writes one long to the EEPROM and waits till done
PRI WriteEEPromLong(addr, value)
  'ser.dbprintf4(string("WriteLong(%d, %8x, %8x, %8x\n"), eeprom_pin, device_sel, addr, value)
  eeprom.WriteLong(eeprom_pin, device_sel, addr, value)
  repeat while(eeprom.WriteWait(eeprom_pin, device_sel, addr))

' This routine locates a free sector
PRI FindFreeSector | sectoraddr
  sectoraddr := first_free
  repeat while (sectoraddr < disk_end)
    if (eeprom.ReadByte(eeprom_pin, device_sel, sectoraddr) & 1)
      first_free := sectoraddr + sector_size
      return sectoraddr
    sectoraddr += sector_size

PRI FreeSectorChain(sectoraddr) | nextaddr
  repeat while (sectoraddr)
    nextaddr := eeprom.ReadLong(eeprom_pin, device_sel, sectoraddr)
    WriteEEPromLong(sectoraddr, nextaddr | 1)
    first_free <#= sectoraddr
    sectoraddr := nextaddr

' This routine searches for a file and returns its starting address
PRI FindFile(fname, dirhand, pfilesize) | firstaddr, filesize, ptr, filename[6]
  if dirhand == 0
    return
  hseek(dirhand, 0)
  ClearCached(dirhand)
  repeat
    if hread(dirhand, @firstaddr, 4) <> 4
      quit
    hread(dirhand, @filesize, 4)
    hread(dirhand, @filename, fname_size)
    if strcomp(fname, @filename) and (firstaddr & attrib_mask)
      long[pfilesize] := filesize
      return firstaddr

PRI remove(fname) | directaddr, sectoraddr, filesize
{{
  This routine removes a file from the files system.  It returns
  a value of zero if the file did not exist, or a one if it did.
}}
  ifnot (sectoraddr := FindFile(fname, @local_handle, @filesize))
    return
  ' Remove file from the directory by clearing it's attribute bits
  sectoraddr &= sector_mask
  directaddr := long[@local_handle + s_sectoraddr] + long[@local_handle + s_offset] - dirent_size
  WriteEEPromLong(directaddr, sectoraddr)
  FreeSectorChain(sectoraddr) ' Free up all the blocks
  return 1
  
' This routine creates a file
PRI CreateFile(fname, modebits) | filesize
  if FindFile(fname, @local_handle, @filesize)
    return
  modebits := (%10000 | modebits) & attrib_mask
  UpdateDirectoryEntry(@local_handle, 0, modebits, 0, fname)
  return 1

PRI OpenRead(fname, handle) | directaddr, firstaddr, filesize
  ifnot (firstaddr := FindFile(fname, @local_handle, @filesize))
    return
  directaddr := long[@local_handle + s_sectoraddr] + long[@local_handle + s_offset] - dirent_size
  InitFileHandle(handle, directaddr, firstaddr, filesize, handle + handle_size)
  return 1

PRI OpenAppend(fname, handle) | directaddr, firstaddr, filesize
  if (firstaddr := FindFile(fname, @local_handle, @filesize))
    directaddr := long[@local_handle + s_sectoraddr] + long[@local_handle + s_offset] - dirent_size
    InitFileHandle(handle, directaddr, firstaddr, filesize, handle + handle_size)
    hseek(handle, filesize)
  else
    UpdateDirectoryEntry(@local_handle, handle, %10110, 0, fname)
  return 1

PRI OpenWrite(fname, handle) | sectoraddr, filesize
  if (sectoraddr := FindFile(fname, @local_handle, @filesize))
    FreeSectorChain(sectoraddr & sector_mask) ' Free up all the blocks
    long[@local_handle + s_position] -= dirent_size
    long[@local_handle + s_offset] -= dirent_size
    UpdateDirectoryEntry(@local_handle, handle, %10110, 0, 0)
  else
    UpdateDirectoryEntry(@local_handle, handle, %10110, 0, fname)
  return 1

PRI UpdateDirectoryEntry(dirhand, handle, firstaddr, filesize, fname) | directaddr
  hwrite(dirhand, @firstaddr, 4)
  hwrite(dirhand, @filesize, 4)
  directaddr := long[dirhand + s_sectoraddr] + long[dirhand + s_offset] - 8
  if fname
    hwrite(dirhand, fname, fname_size)
  hclose(dirhand) ' Force writes to happen
  if handle
    InitFileHandle(handle, directaddr, firstaddr, filesize, handle + handle_size) ' Should specify buffer
